import numpy
import pytest

from darfix.core.imageStack import FixedDimension
from darfix.core.imageStack import ImageStack


def flatten_scans(data: numpy.ndarray):
    new_shape = (
        data.size // data.shape[-2] // data.shape[-1],
        data.shape[-2],
        data.shape[-1],
    )
    return data.reshape(new_shape)


def legacy_get_data(
    dataset: ImageStack, indices=None, dimension=None, return_indices=False
):
    """
    Legacy very complex dataset function
    (With a few modifications to remove ImageDataset dependencies)
    """
    if dimension is not None and len(dataset._data.shape) > 3:
        # Make sure dimension and value are lists
        if isinstance(dimension[0], int):
            dimension[0] = [dimension[0]]
            dimension[1] = [dimension[1]]
        data = dataset.data

        # Init list of bool indices
        bool_indices = numpy.zeros(dataset.nframes, dtype=bool)
        if indices is None:
            indices = numpy.arange(dataset.nframes)
        bool_indices[indices] = True
        bool_indices = bool_indices.reshape(dataset.scan_shape)
        indx = numpy.arange(dataset.nframes).reshape(dataset.scan_shape)

        # For every axis, get corresponding elements
        for i, dim in enumerate(sorted(dimension[0])):
            # Flip axis to be consistent with the data shape
            axis = len(dataset.scan_shape) - dim - 1
            data = data.take(indices=dimension[1][i], axis=axis)
            bool_indices = bool_indices.take(indices=dimension[1][i], axis=axis)
            indx = indx.take(indices=dimension[1][i], axis=axis)

        data = data[bool_indices]
        indx = indx[bool_indices]
        if return_indices:
            return flatten_scans(data), indx.flatten()
        return flatten_scans(data)

    data = flatten_scans(dataset.data)
    if return_indices:
        if indices is None:
            indices = numpy.arange(dataset.nframes)
        return data[indices], indices
    if indices is None:
        return data
    return data[indices]


@pytest.fixture
def image_stack():
    scans = numpy.arange(2 * 2).repeat(5 * 4 * 3).reshape(5, 4, 3, 2, 2)
    return ImageStack(scans)


def test_properties(image_stack: ImageStack):
    assert image_stack.nframes == 5 * 4 * 3
    assert image_stack.frame_shape == (2, 2)
    assert image_stack.scan_shape == (5, 4, 3)


def test_filter_indices(image_stack: ImageStack):
    numpy.testing.assert_array_equal(image_stack.filter_indices([0, 1, 2]), [0, 1, 2])
    numpy.testing.assert_array_equal(image_stack.filter_indices(None), numpy.arange(60))

    numpy.testing.assert_array_equal(
        image_stack.filter_indices(fixed_dimension=FixedDimension(0, 0)),
        numpy.arange(0, 60, 3),
    )
    numpy.testing.assert_array_equal(
        image_stack.filter_indices(fixed_dimension=FixedDimension(0, 1)),
        numpy.arange(1, 60, 3),
    )
    numpy.testing.assert_array_equal(
        image_stack.filter_indices(fixed_dimension=FixedDimension(2, 1)),
        numpy.arange(12, 24),
    )

    numpy.testing.assert_array_equal(
        image_stack.filter_indices(
            fixed_dimension=FixedDimension(2, 1), indices=numpy.arange(0, 18)
        ),
        numpy.arange(12, 18),
    )


def test_get_filtered_data(image_stack: ImageStack):
    # Should give the same result as old ImageDataset::get_data

    numpy.testing.assert_array_equal(
        image_stack.get_filtered_data([0, 1, 2]),
        legacy_get_data(image_stack, [0, 1, 2]),
    )
    numpy.testing.assert_array_equal(
        image_stack.get_filtered_data(None), legacy_get_data(image_stack, None)
    )

    numpy.testing.assert_array_equal(
        image_stack.get_filtered_data(fixed_dimension=FixedDimension(0, 0)),
        legacy_get_data(image_stack, dimension=[[0], [0]]),
    )
    numpy.testing.assert_array_equal(
        image_stack.get_filtered_data(fixed_dimension=FixedDimension(0, 1)),
        legacy_get_data(image_stack, dimension=[[0], [1]]),
    )
    numpy.testing.assert_array_equal(
        image_stack.get_filtered_data(fixed_dimension=FixedDimension(2, 1)),
        legacy_get_data(image_stack, dimension=[[2], [1]]),
    )

    numpy.testing.assert_array_equal(
        image_stack.get_filtered_data(
            fixed_dimension=FixedDimension(2, 1), indices=numpy.arange(0, 18)
        ),
        legacy_get_data(image_stack, dimension=[[2], [1]], indices=numpy.arange(0, 18)),
    )



import os
import time

from ewoksorange.tests.conftest import qtapp  # noqa F401
from silx.io.url import DataUrl

import darfix.resources.tests
from darfix.core.dataset import ImageDataset
from darfix.core.dimension import Dimension
from darfix.dtypes import Dataset
from orangecontrib.darfix.widgets.transformation import TransformationWidgetOW


def test_TransformationWidgetOW(
    resource_files,
    qtapp,  # noqa F811
):
    _316H_dummy_insitu_g1_RSM_2 = str(
        resource_files(darfix.resources.tests).joinpath(
            os.path.join("transformation", "316H_dummy_insitu_g1_RSM_2.h5")
        )
    )

    data_file_url = DataUrl(
        file_path=_316H_dummy_insitu_g1_RSM_2,
        data_path="/1.1/measurement/basler_ff",
        scheme="silx",
    )

    dataset = Dataset(
        dataset=ImageDataset(
            detector_url=data_file_url.path(),
            metadata_url=DataUrl(
                file_path=str(_316H_dummy_insitu_g1_RSM_2),
                data_path="/1.1/instrument/positioners",
                scheme="silx",
            ).path(),
            _dir=None,
        ),
    )

    # Actually the dataset 316H_dummy_insitu_g1_RSM_2 is weird as diffry size does not match number of frames
    dataset.dataset.dims.add_dim(0, Dimension("diffry", 11))

    widget = TransformationWidgetOW()
    widget.setDataset(dataset)
    assert dataset.dataset.transformation is None
    widget._execute_task()
    # Wait until the task finished
    time.sleep(1)
    new_dataset: Dataset = widget.get_task_output_value("dataset")
    assert new_dataset.dataset.transformation is not None
    assert new_dataset.dataset.transformation.kind == "magnification"

    widget._methodCB.setCurrentText("RSM")
    widget._execute_task()
    # Wait until the task finished
    time.sleep(1)
    new_dataset: Dataset = widget.get_task_output_value("dataset")
    assert new_dataset.dataset.transformation is not None
    assert new_dataset.dataset.transformation.kind == "rsm"



from __future__ import annotations

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix import dtypes
from darfix.gui.rocking_curves.rockingCurvesWidget import RockingCurvesWidget
from darfix.tasks.rocking_curves import RockingCurves


class RockingCurvesWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=RockingCurves):
    name = "rocking curves"
    icon = "icons/curves.png"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("int_thresh", "method", "output_filename")

    def __init__(self):
        super().__init__()

        self._widget = RockingCurvesWidget(parent=self)
        self._widget.sigFitClicked.connect(self._launch_fit)
        self.mainArea.layout().addWidget(self._widget)
        int_thresh = self.get_task_input_value("int_thresh", None)
        if int_thresh is not None:
            self._widget.setIntensityThreshold(int_thresh)
        method = self.get_task_input_value("method", None)
        if method is not None:
            self._widget.setFitMethod(method)

    def handleNewSignals(self) -> None:
        dataset = self.get_task_input_value("dataset")
        self.setDataset(dataset, pop_up=True)

        # Do not call super().handleNewSignals() to prevent propagation

    def setDataset(self, dataset: dtypes.Dataset | None, pop_up=True):
        if dataset is None:
            return
        self._widget.setDataset(dataset)
        if pop_up:
            self.open()

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()

        # Saving is handled by the widget
        task_inputs["output_filename"] = None

        return task_inputs

    def task_output_changed(self) -> None:
        self._widget.onFitFinished()
        dataset = self.get_task_output_value("dataset", None)
        maps = self.get_task_output_value("maps", None)
        if dataset is not None and maps is not None:
            if not isinstance(dataset, dtypes.Dataset):
                raise dtypes.DatasetTypeError(dataset)
            self._widget.updateDataset(dataset.dataset, maps)

    def _launch_fit(self):
        self.set_default_input("int_thresh", self._widget.getIntensityThreshold())
        self.set_default_input("method", self._widget.getFitMethod())
        self.execute_ewoks_task()



import os

import h5py
import numpy
from ewoksorange.tests.conftest import qtapp  # noqa F401
from silx.io.url import DataUrl

from orangecontrib.darfix.widgets.dimensions import DimensionWidgetOW

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files

import darfix.resources.tests
from darfix.dtypes import Dataset
from darfix.dtypes import ImageDataset


def test_dimension_NiTi_1PD_002_g411_420MPa_mosalayers_2x(tmp_path, qtapp):  # noqa F811
    """
    test dimension search with a 'real use case' motor positions
    """
    dataset_file = resource_files(darfix.resources.tests).joinpath(
        os.path.join(
            "dimensions_definition", "NiTi_1PD_002_g411_420MPa_mosalayers_2x.h5"
        )
    )

    raw_motor_values = {}
    with h5py.File(dataset_file, mode="r") as h5f:
        raw_motor_values["chi"] = h5f["positioners/chi"][()]
        raw_motor_values["diffry"] = h5f["positioners/diffry"][()]
        raw_motor_values["difftz"] = h5f["positioners/difftz"][()]

    data_folder = tmp_path / "test_fitting"
    data_folder.mkdir()
    data_file_url = DataUrl(
        file_path=os.path.join(str(data_folder), "data.h5"),
        data_path="data",
        scheme="silx",
    )
    number_of_points = 31500
    with h5py.File(data_file_url.file_path(), mode="w") as h5f:
        h5f["data"] = numpy.random.random(number_of_points)

    dataset = Dataset(
        dataset=ImageDataset(
            detector_url=data_file_url.path(),
            metadata_url=DataUrl(
                file_path=str(dataset_file),
                data_path="positioners",
                scheme="silx",
            ).path(),
            _dir=None,
        )
    )

    widget = DimensionWidgetOW()

    def find_dimensions_for_tolerance(tolerance: float):
        widget._widget.setTolerance(tolerance=tolerance)
        widget._findDimensions()

    def check_dimensions_bounds(tolerance):
        """Make sure find_dimension is correctly fitting motor bounds"""
        # check dims object vs dataset real value
        for dim in widget._widget.dims.values():
            dim = dim.toDimension()
            numpy.testing.assert_almost_equal(
                dim.start, min(raw_motor_values[dim.name]), decimal=4
            )
            numpy.testing.assert_almost_equal(
                dim.stop, max(raw_motor_values[dim.name]), decimal=4
            )

        # check display (DimensionItem instances) vs real value
        for dim_item in widget._widget.dims.values():
            numpy.testing.assert_almost_equal(
                dim_item.start, min(raw_motor_values[dim_item.name]), decimal=4
            )
            numpy.testing.assert_almost_equal(
                dim_item.stop, max(raw_motor_values[dim_item.name]), decimal=4
            )

    widget.setDataset(dataset=dataset)
    # check if tolerance == 1e-5
    find_dimensions_for_tolerance(tolerance=1e-5)

    check_dimensions_bounds(tolerance=1e-5)

    # check if tolerance == 1e-4
    find_dimensions_for_tolerance(tolerance=1e-4)
    check_dimensions_bounds(tolerance=1e-4)



import numpy
import pytest
from silx.opencl.common import ocl
from skimage import data

from darfix.core.componentsMatching import ComponentsMatching
from darfix.core.componentsMatching import Method


@pytest.fixture(scope="module")
def components_matching():
    moon = data.moon()
    camera = data.camera()
    gravel = data.gravel()
    components1 = numpy.array([moon, camera, gravel])
    components2 = numpy.array([gravel, moon, camera])

    return ComponentsMatching(components=(components1, components2))


def test_euclidean_distance(components_matching):
    components1, components2 = components_matching.components

    assert components_matching.euclidean_distance(components1[0], components2[1]) == 0
    assert components_matching.euclidean_distance(components1[1], components2[1]) != 0


@pytest.mark.skipif(ocl is None, reason="PyOpenCl is missing")
def test_sift_match(components_matching):
    final_matches, matches = components_matching.match_components(
        method=Method.sift_feature_matching
    )
    assert final_matches[0] == 1


def test_orb_match(components_matching):
    final_matches, matches = components_matching.match_components(
        method=Method.orb_feature_matching
    )
    assert final_matches[0] == 1


def test_draw_matches0(components_matching):
    final_matches, matches = components_matching.match_components(
        method=Method.orb_feature_matching
    )
    stack = components_matching.draw_matches(
        final_matches, matches, displayMatches=True
    )
    assert stack[0].shape == (512, 1024)
    stack = components_matching.draw_matches(
        final_matches, matches, displayMatches=False
    )
    assert stack[1].shape == (512, 1024)


def test_draw_matches1(components_matching):
    final_matches, matches = components_matching.match_components(
        method=Method.euclidean_distance
    )
    stack = components_matching.draw_matches(final_matches, matches)
    assert stack[2].shape == (512, 1024)


@pytest.mark.skipif(ocl is None, reason="PyOpenCl is missing")
def test_draw_matches2(components_matching):
    final_matches, matches = components_matching.match_components(
        method=Method.sift_feature_matching
    )
    stack = components_matching.draw_matches(final_matches, matches)
    assert stack[2].shape == (512, 1024)



from ewokscore import Task


class MetadataTask(
    Task,
    input_names=[
        "dataset",
    ],
):
    """
    'Place holder' task for the MetadataWidgetOW.
    This widget only does display of information an no processing.
    But having this task allow use to have it compatible with ewoks.
    """

    def run(self):
        pass



from typing import Tuple

import numpy
from scipy.stats import multivariate_normal

Vector3D = Tuple[float, float, float]


def sturges_rules(n: int) -> int:
    """https://en.wikipedia.org/wiki/Sturges%27s_rule"""
    return int(numpy.ceil(1 + numpy.log2(n)))


def gaussian(
    x: numpy.ndarray, amplitude: float, x0: float, std_dev: float, background: float
):
    """
    Gaussian function (https://en.wikipedia.org/wiki/Gaussian_function) with background

    :param x: value where to evaluate
    :param amplitude: peak height
    :param x0: peak center
    :param std_dev: standard deviation
    :param background: lowest value of the curve (value of the limits)

    :returns: result of the function on x
    :rtype: float
    """
    return background + amplitude * numpy.exp(
        -numpy.power(x - x0, 2) / (2 * numpy.power(std_dev, 2))
    )


def bivariate_gaussian(
    X: numpy.ndarray,
    x0_0: float,
    x1_0: float,
    sigma_x0: float,
    sigma_x1: float,
    amplitude: float,
    correlation: float = 0,
    background: float = 0,
):
    """
    Bivariate case of the gaussian function with background (https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Bivariate_case)

    If correlation is set to 0 (default), this is simply a 2D gaussian

    :param x: value where to evaluate
    """
    x0, x1 = X

    return background + amplitude * numpy.exp(
        -0.5
        / (1 - correlation**2)
        * (
            ((x0 - x0_0) / sigma_x0) ** 2
            + ((x1 - x1_0) / sigma_x1) ** 2
            - 2 * correlation * (x0 - x0_0) * (x1 - x1_0) / sigma_x0 / sigma_x1
        )
    )


def trivariate_gaussian(
    X: numpy.ndarray,
    x0_0: float,
    x1_0: float,
    x2_0: float,
    sigma_x0: float,
    sigma_x1: float,
    sigma_x2: float,
    c10: float,
    c12: float,
    c20: float,
    amplitude: float,
    background: float = 0,
):
    """
    Trivariate case of the gaussian function with background (https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Density_function)

    :param X: stack of N vectors of shape (3, N)
    :param x0_0: mean along axis 0
    :param x1_0: mean along axis 1
    :param x2_0: mean along axis 2
    :param sigma_x0: standard deviation along axis 0
    :param sigma_x1: standard deviation along axis 1
    :param sigma_x2: standard deviation along axis 2
    :param c10: cross-correlation factor between axis 0 and axis 1
    :param c12: cross-correlation factor between axis 1 and axis 2
    :param c02: cross-correlation factor between axis 0 and axis 2
    :param amplitude:
    :param background:
    """
    assert X.shape[0] == 3

    covariance_matrix = numpy.array(
        [
            [sigma_x0**2, c10 * sigma_x0 * sigma_x1, c20 * sigma_x0 * sigma_x2],
            [c10 * sigma_x0 * sigma_x1, sigma_x1**2, c12 * sigma_x1 * sigma_x2],
            [c20 * sigma_x0 * sigma_x2, c12 * sigma_x1 * sigma_x2, sigma_x2**2],
        ]
    )

    X0 = numpy.array([x0_0, x1_0, x2_0])

    return background + amplitude * multivariate_normal.pdf(
        X.T, X0, covariance_matrix, allow_singular=True
    )



from __future__ import annotations

import numpy
import tqdm

from darfix.core.dataset import ImageDataset
from darfix.core.dimension import AcquisitionDims
from darfix.core.dimension import Dimension


def reorder_frames_of_zigzag_scan(
    dims: AcquisitionDims,
    dataset: ImageDataset,
) -> None:
    """
    Modify data to reorder frames

    Prequisites : user already check that ZIGZAG mode is used for acquisition.
    """

    if dims.ndim <= 1:
        raise NotImplementedError("Zigzag mode is not available for 1D dataset")

    indices = numpy.arange(dataset.nframes)

    # Take first dim -> fast motor
    fast_motor = tuple(dims.values())[0]

    assert isinstance(fast_motor, Dimension)

    # Reorder indices : In ZIGZAG mode fast motor is moving from start to stop then from stop to start value N times, with N corresponding to data.nframes // fast_motor.size.
    # When motor is moving backward, we reverse the indices of data to have the same order of frames as in normal acquisition mode.

    fast_motor_continuous_movement_count = dataset.nframes // fast_motor.size

    indices = indices.reshape(fast_motor_continuous_movement_count, fast_motor.size)
    indices[1::2] = indices[1::2, ::-1]
    indices = indices.flatten()

    for key in dataset.metadata_dict.keys():
        dataset.metadata_dict[key] = dataset.metadata_dict[key][indices]

    # quite the same but i use a loop and tqdm for better user experience
    for i, chunk in tqdm.tqdm(
        enumerate(
            numpy.split(dataset.as_array3d(), fast_motor_continuous_movement_count)
        ),
        "reorder frames for zigzag scan...",
        fast_motor_continuous_movement_count,
    ):
        if i % 2 == 0:
            continue
        chunk[:] = chunk[::-1]



from __future__ import annotations

import numpy
from silx.gui import qt

from darfix import dtypes
from darfix.gui.shiftcorrection.shiftCorrectionWidget import ShiftCorrectionWidget
from darfix.tasks.shiftcorrection import ShiftCorrection
from orangecontrib.darfix.widgets.operationBase import OperationWidgetBase


class ShiftCorrectionWidgetOW(OperationWidgetBase, ewokstaskclass=ShiftCorrection):
    """
    Widget to make the shift correction of a dataset.
    """

    name = "shift correction"
    description = "A widget to perform shift correction"
    icon = "icons/shift_correction.svg"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("shift", "selected_axis", "selected_index")

    def __init__(self):
        super().__init__()
        qt.QLocale.setDefault(qt.QLocale("en_US"))

    def createMainWidget(self, default_inputs: dict) -> ShiftCorrectionWidget:
        widget = ShiftCorrectionWidget()
        widget.correctSignal.connect(self.execute_shift_correction)
        widget.setCorrectionInputs(
            default_inputs.get("shift", (0, 0)),
            default_inputs.get("selected_axis", None),
        )
        return widget

    def saveMainWidget(self) -> dict:
        inputs = self.mainWidget.getCorrectionInputs()
        for key, value in inputs.items():
            if isinstance(value, numpy.ndarray):
                value = value.tolist()
            self.set_default_input(key, value)

    @property
    def mainWidget(self) -> ShiftCorrectionWidget:
        return super().mainWidget

    def handleNewSignals(self) -> None:
        super().handleNewSignals()
        dataset = self.get_task_input_value("dataset", None)
        if dataset is not None:
            self.setDataset(dataset, pop_up=True)

    def setDataset(self, dataset: dtypes.Dataset, pop_up=True):
        self.mainWidget.setDataset(dataset)
        if pop_up:
            self.open()

    def execute_shift_correction(self):
        self.saveMainWidget()
        self.execute_ewoks_task_without_propagation()

    def task_output_changed(self) -> None:
        super().task_output_changed()
        self.mainWidget.refreshPlot()



import logging
from typing import Optional

from ewokscore.missing_data import MISSING_DATA
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from ewoksorange.gui.parameterform import block_signals

from darfix import dtypes
from darfix.gui.operationThread import OperationThread
from darfix.gui.utils.message import missing_dataset_msg
from darfix.gui.weakBeamWidget import WeakBeamWidget
from darfix.tasks.weakbeam import WeakBeam

_logger = logging.getLogger(__name__)


class WeakBeamWidgetOW(
    OWEwoksWidgetOneThread,
    ewokstaskclass=WeakBeam,
):
    """
    Widget that computes dataset with filtered weak beam and recover its Center of Mass.
    """

    name = "weak beam"
    icon = "icons/gaussian.png"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("nvalue", "indices", "title")

    def __init__(self):
        super().__init__()
        self._thread = None

        self._widget = WeakBeamWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)
        nvalue = self.get_default_input_value("nvalue", MISSING_DATA)
        if nvalue is not MISSING_DATA:
            with block_signals(self._widget):
                self._widget.nvalue = nvalue
        else:
            self.set_default_input("nvalue", self._widget.nvalue)

        # connect signal / slot
        self._widget.sigValidate.connect(self.execute_ewoks_task)
        self._widget.sigApplyThreshold.connect(self._launch_recover_weak_beam)
        self._widget.sigNValueChanged.connect(self._nValueChanged)

    def setDataset(self, dataset: Optional[dtypes.Dataset], pop_up: bool = False):
        if dataset is None:
            return
        if pop_up:
            self.open()
        self.set_dynamic_input("dataset", dataset=dataset)
        # if some processing on-going with the previous dataset stop it.
        self._stop_thread()

    def _nValueChanged(self):
        self.set_default_input("nvalue", self._widget.nvalue)

    def handleNewSignals(self) -> None:
        dataset = self.get_task_input_value("dataset", None)
        if dataset is None:
            return

        self.setDataset(dataset=dataset, pop_up=True)

        if isinstance(dataset, dtypes.Dataset):
            self.set_dynamic_input("indices", dataset.indices)
            title = dataset.dataset.title
            self.set_default_input(
                "title",
                title if title is not None else MISSING_DATA,
            )

        # warning: do not call to make sure the processing is not triggered
        # return super().handleNewSignals()

    def task_output_changed(self) -> None:
        dataset = self.get_task_output_value("dataset", MISSING_DATA)
        self.setDataset(dataset)

    def _launch_recover_weak_beam(self):
        """callback when the user modify the 'nvalue'"""
        if self._thread is not None:
            _logger.warning("recover weak beam already on going")

        dataset = self.get_task_input_value("dataset", None)
        if dataset is None:
            missing_dataset_msg()
            return
        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)

        self._widget.setProcessingButtonsEnabled(False)
        nvalue = self.get_task_input_value("nvalue", MISSING_DATA)
        # FIXME: all processing should be in done in a ewoks task. See https://gitlab.esrf.fr/XRD/darfix/-/issues/130
        self._thread = OperationThread(self, dataset.dataset.recover_weak_beam)
        self._thread.setArgs(
            n=nvalue,
            indices=dataset.indices,
        )
        self._thread.finished.connect(self._recover_weak_beam_finished)
        self._thread.start()

    def _stop_thread(self):
        if self._thread is not None:
            self._thread.finished.disconnect(self._recover_weak_beam_finished)
        self._thread = None
        self._widget.setProcessingButtonsEnabled(True)

    def _recover_weak_beam_finished(self):
        sender = self.sender()
        dataset = self._thread.data
        self._stop_thread()

        if dataset in (None, MISSING_DATA):
            return
        assert isinstance(
            dataset, dtypes.ImageDataset
        ), f"dataset is expected to be an instance of {dtypes.ImageDataset}. Get {type(dataset)}"
        self._widget.setResult(
            center_of_mass=dataset.apply_moments(indices=sender.kwargs["indices"])[0][
                0
            ],
            transformation=dataset.transformation,
        )
        self._stop_thread()



import pytest
from silx.io.url import DataUrl
from silx.resources import ExternalResources

from ..core.data_selection import load_process_data
from . import utils


@pytest.fixture(scope="session")
def resource_files():
    try:
        from importlib.resources import files
    except ImportError:
        from importlib_resources import files

    return files


@pytest.fixture
def dataset(tmpdir):

    return utils.create_3motors_dataset(tmpdir)


@pytest.fixture
def input_dataset(tmp_path, silx_resources):
    input_filename = silx_resources.getfile("input.h5")

    detector_url = DataUrl(
        file_path=input_filename, data_path="/2.1/measurement/my_detector"
    )
    metadata_url = DataUrl(
        file_path=input_filename, data_path="/2.1/instrument/positioners"
    )

    dataset, _, _, _ = load_process_data(
        detector_url=detector_url,
        root_dir=tmp_path,
        title="input",
        metadata_url=metadata_url,
    )
    return dataset


@pytest.fixture(scope="session")
def silx_resources(tmp_path_factory):
    silx_resources = ExternalResources(
        "darfix",
        url_base="http://www.silx.org/pub/darfix",
        # Create a sub-folder to avoid name collision
        data_home=tmp_path_factory.mktemp("external"),
    )
    yield silx_resources
    silx_resources.clean_up()



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "12/04/2022"

import numpy
from silx.gui import icons
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot1D
from silx.gui.plot import ScatterView
from silx.gui.plot import StackView

import darfix
from darfix.io.utils import write_components

from .chooseDimensions import ChooseDimensionDock
from .utils.utils import select_output_hdf5_file_with_dialog


class DisplayComponentsWidget(qt.QMainWindow):
    """
    Widget to display rocking curves and scattering plot from a set of components
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # Widget with the type of bss and the number of components to compute
        widget = qt.QWidget(self)
        self._sv_components = StackView(parent=self)
        self._sv_components.setColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        self._sv_components.setKeepDataAspectRatio(True)
        self._sv_components.setGraphTitle("Components")
        self._mixing_plots_widget = MixingPlotsWidget(self)
        self._mixing_plots_widget.activeCurveChanged.connect(self._setComponent)
        self._sv_components.sigFrameChanged.connect(self._mixing_plots_widget._setCurve)

        self.bottom_widget = qt.QWidget(self)
        layout = qt.QGridLayout()
        componentsLabel = qt.QLabel("Components")
        rockingCurvesLabel = qt.QLabel("Rocking curves")
        font = qt.QFont()
        font.setBold(True)
        componentsLabel.setFont(font)
        rockingCurvesLabel.setFont(font)
        rockingCurvesLabel.setFont(font)
        layout.addWidget(componentsLabel, 1, 0, 1, 2, qt.Qt.AlignCenter)
        layout.addWidget(rockingCurvesLabel, 1, 2, 1, 2, qt.Qt.AlignCenter)
        layout.addWidget(self._sv_components, 2, 0, 2, 2)
        layout.addWidget(self._mixing_plots_widget, 2, 3, 2, 2)
        self.saveB = qt.QPushButton("Save components")
        self.saveB.pressed.connect(self._saveComp)
        layout.addWidget(self.saveB, 4, 4, 1, -1)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def hideButton(self):
        self._computeB.hide()

    def showButton(self):
        self._computeB.show()

    def setComponents(self, components, W, dimensions, values, title=""):
        """
        Components setter. Updates the plots with the components and their
        corresponding rocking curves.

        :param array_like components: stack of images with the components
        :param array_like W: array with the rocking curves intensity
        :param dict dimensions: dictionary with the values of the dimensions
        """
        self.components = numpy.round(components, 5)
        self.W = W
        self.values = values
        self.dimensions = dimensions
        self._sv_components.setStack(self.components)
        if title != "":
            self._sv_components.setTitleCallback(lambda idx: title)
        self._mixing_plots_widget.updatePlots(self.dimensions, self.W.T, values)

    def _saveComp(self):
        """Save components to a HDF5 file"""
        filename = select_output_hdf5_file_with_dialog()
        if filename:
            write_components(
                filename,
                "entry",
                self.dimensions.to_dict(),
                self.W,
                self.components,
                self.values,
                1,
            )

    def _setComponent(self, index=None):
        if index is not None:
            status = self._sv_components.blockSignals(True)
            self._sv_components.setFrameNumber(index)
            self._sv_components.blockSignals(status)

    def setStackViewColormap(self, colormap):
        """
        Sets the stackView colormap

        :param colormap: Colormap to set
        :type colormap: silx.gui.colors.Colormap
        """
        self._sv_components.setColormap(colormap)

    def getStackViewColormap(self):
        """
        Returns the colormap from the stackView

        :rtype: silx.gui.colors.Colormap
        """
        return self._sv_components.getColormap()


class MixingPlotsWidget(qt.QMainWindow):
    """
    Widget to show rocking curves and reciprocal space maps based on the mixing matrix.
    """

    activeCurveChanged = qt.Signal(int)

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # Widget with the type of bss and the number of components to compute
        widget = qt.QWidget(self)

        self._is2dimensional = False

        self._plot_rocking_curves = PlotRockingCurves()
        self._plot_rocking_curves.sigStateDisabled.connect(self._wholeStack)
        self._plot_rocking_curves.sigMotorChanged.connect(self._updateMotorAxis)
        self._plot_rocking_curves.sigActiveCurveChanged.connect(
            self._activeCurveChanged
        )

        self._rsm_scatter = ScatterView()
        self._rsm_scatter.hide()

        self._toolbar = qt.QToolBar(parent=self)
        self._toolbar.setIconSize(self._toolbar.iconSize() * 1.2)
        curves_icon = icons.getQIcon("darfix:gui/icons/curves")
        scatter_icon = icons.getQIcon("darfix:gui/icons/scatter")
        self.curves_action = qt.QAction(curves_icon, "Curves", self)
        self.curves_action.setCheckable(True)
        self.curves_action.setChecked(True)
        self.rsm_action = qt.QAction(scatter_icon, "Scatter", self)
        self.rsm_action.setCheckable(True)
        self.curves_action.toggled.connect(self._activateCurvesPlot)
        self.rsm_action.triggered.connect(self.curves_action.toggle)
        self._toolbar.addAction(self.curves_action)
        self._toolbar.addAction(self.rsm_action)
        self._toolbar.setOrientation(qt.Qt.Vertical)
        self._toolbar.hide()

        self.bottom_widget = qt.QWidget(self)
        layout = qt.QGridLayout()
        layout.addWidget(self._plot_rocking_curves, 1, 0, 2, 2)
        layout.addWidget(self._rsm_scatter, 1, 0, 2, 2)
        layout.addWidget(self._toolbar, 1, 3, 2, 1)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def is2dimensional(self, is2dimensional):
        self._is2dimensional = is2dimensional

    @property
    def scatter(self):
        return self._rsm_scatter

    def updatePlots(self, dimensions, mixing_matrix, dimensions_values):
        self.dimensions = dimensions
        self.mixing_matrix = mixing_matrix
        self.is2dimensional(True if dimensions.ndim == 2 else False)
        self._activateCurvesPlot(True)
        self.clearPlots()
        self.values = numpy.array(list(dimensions_values.values()))

        for i, row in enumerate(self.mixing_matrix):
            self._plot_rocking_curves.getPlot().addCurve(
                numpy.arange(len(row)), row, legend=str(i)
            )

        if self._is2dimensional:
            colormap = Colormap(name="jet", normalization="linear")
            self._rsm_scatter.getPlotWidget().setGraphXLabel(
                self.dimensions.get(0).name
            )
            self._rsm_scatter.getPlotWidget().setGraphYLabel(
                self.dimensions.get(1).name
            )
            self._rsm_scatter.setData(
                self.values[0], self.values[1], self.mixing_matrix[0]
            )
            self._rsm_scatter.setColormap(colormap)
            self._rsm_scatter.resetZoom()
            self._toolbar.show()
        else:
            self._toolbar.hide()

        if self.dimensions.ndim > 1:
            # Motor dimension and index
            self.dimension = 0
            self.index = 0
            self._plot_rocking_curves._chooseDimensionDock.show()
            self._plot_rocking_curves._chooseDimensionDock.widget().setDimensions(
                self.dimensions
            )
        else:
            self._plot_rocking_curves._chooseDimensionDock.hide()

        self._plot_rocking_curves.getPlot().setActiveCurve("0")

    def _activateCurvesPlot(self, checked=False):
        if checked:
            self.rsm_action.setChecked(False)
            self._rsm_scatter.hide()
            self._plot_rocking_curves.show()
        else:
            self._plot_rocking_curves.hide()
            self._rsm_scatter.show()

    def _activeCurveChanged(self, prev_legend=None, legend=None):
        if legend:
            self.activeCurveChanged.emit(int(legend))
            if self._is2dimensional:
                self._rsm_scatter.setData(
                    self.values[0], self.values[1], self.mixing_matrix[int(legend)]
                )

    def _setCurve(self, index=-1):
        if index >= 0:
            status = self._plot_rocking_curves.blockSignals(True)
            self._plot_rocking_curves.getPlot().setActiveCurve(str(index))
            self._plot_rocking_curves.blockSignals(status)
            if self._is2dimensional:
                self._rsm_scatter.setData(
                    self.values[0], self.values[1], self.mixing_matrix[index]
                )

    def _updateFrameNumber(self, index):
        """Update the current plot.

        :param index: index of the frame to be displayed
        """
        self.index = index
        self.clearPlots()
        dim_values = numpy.take(
            self.values.reshape(self.dimensions.shape), self.index, self.dimension
        )
        for i, row in enumerate(self.mixing_matrix):
            W = numpy.take(
                row.reshape(self.dimensions.shape), self.index, self.dimension
            )
            self._plot_rocking_curves.addCurve(dim_values, W, legend=str(i))

    def _wholeStack(self):
        # Whole rocking curves are showed
        self._plot_rocking_curves.clear()
        for i, row in enumerate(self.mixing_matrix):
            self._plot_rocking_curves.addCurve(
                numpy.arange(len(row)), row, legend=str(i)
            )
        self._plot_rocking_curves.getPlot().setGraphXLabel("Image id")

    def _updateMotorAxis(self, dim=0, val=0):
        """
        Updates the motor to show the rocking curve from.

        :param int axis: 0 if no motor is chosen, else axis of the motor.
        """
        dimension = [dim, val]
        # Make sure dimension and value are lists
        if type(dimension[0]) is int:
            dimension[0] = [dimension[0]]
            dimension[1] = [dimension[1]]
        indx = numpy.arange(len(self.mixing_matrix[0])).reshape(self.dimensions.shape)
        # For every axis, get corresponding elements
        for i, j in sorted(zip(dimension[0], dimension[1])):
            # Flip axis to be consistent with the data shape
            axis = self.dimensions.ndim - i - 1
            # dim_values = numpy.array(list(self.dimensions.values())[dim])
            indx = indx.take(indices=j, axis=axis)
        self._plot_rocking_curves.clear()
        axis = 0
        for i in range(self.dimensions.ndim):
            if i not in dimension[0]:
                axis = i
        text = self.dimensions.get(axis).name
        self._plot_rocking_curves.getPlot().setGraphXLabel(text)
        if len(indx.shape) == 1:
            for i, row in enumerate(self.mixing_matrix):
                W = row[indx]
                self._plot_rocking_curves.addCurve(
                    self.values[axis][indx], W, legend=str(i)
                )

    def clearPlots(self):
        self._plot_rocking_curves.clear()


class PlotRockingCurves(qt.QMainWindow):
    """
    Widget to plot the rocking curves of the components. It can be filtered
    to show only the rocking curves of a certain moving dimension.
    """

    sigStateDisabled = qt.Signal()
    sigMotorChanged = qt.Signal(list, list)
    sigActiveCurveChanged = qt.Signal(object, object)

    def __init__(self, parent=None, dimensions=None):
        qt.QMainWindow.__init__(self, parent)

        self._plot = Plot1D()
        self._plot.setGraphTitle("Rocking curves")
        self._plot.setGraphXLabel("Image id")
        self._plot.sigActiveCurveChanged.connect(self.sigActiveCurveChanged)

        self._chooseDimensionDock = ChooseDimensionDock(self, vertical=False)
        self._chooseDimensionDock.hide()
        self._motors = qt.QWidget()
        self._motors.hide()
        layout = qt.QVBoxLayout()
        layout.addWidget(self._plot)

        centralWidget = qt.QWidget()
        centralWidget.setLayout(layout)
        self.setCentralWidget(centralWidget)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._chooseDimensionDock)

        self.addCurve = self._plot.addCurve
        self._chooseDimensionDock.widget().filterChanged.connect(self.sigMotorChanged)
        self._chooseDimensionDock.widget().stateDisabled.connect(self.sigStateDisabled)

    def clear(self):
        self._plot.clear()

    def getPlot(self):
        return self._plot



from __future__ import annotations

import os.path
from pathlib import Path
from typing import Literal

from ewokscore import Task
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from silx.io.dictdump import dicttonx

from ..core.rocking_curves import compute_residuals
from ..core.rocking_curves import generate_rocking_curves_nxdict
from ..dtypes import Dataset


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """ Input dataset containing a stack of images """
    int_thresh: float | None = None
    """If provided, only the rocking curves with higher ptp (peak to peak) value > int_thresh are fitted, others are assumed to be noise and will be discarded"""
    method: Literal["trf", "lm", "dogbox"] = "trf"
    """Method to use for the rocking curves fit"""
    output_filename: str | Path | None = None
    """Output filename to save the rocking curves results. Results are not saved if not provided"""


class RockingCurves(Task, input_model=Inputs, output_names=["dataset", "maps"]):
    """Analyze the rocking curve of each pixel of each image of the darfix dataset by fitting to a peak shape, e.g. a Gaussian.

    Related article : https://pmc.ncbi.nlm.nih.gov/articles/PMC10161887/#sec3.3.1
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())

        output_filename = inputs.output_filename
        if output_filename and os.path.isfile(output_filename):
            raise FileExistsError(
                f"""Cannot launch rocking curves fit: saving destination {output_filename} already exists.
                Change the `output_filename` input or set it to None to disable saving."""
            )

        dataset = inputs.dataset.dataset
        indices = inputs.dataset.indices
        new_image_dataset, maps = dataset.apply_fit(
            indices=indices, int_thresh=inputs.int_thresh, method=inputs.method
        )

        if output_filename is not None:
            nxdict = generate_rocking_curves_nxdict(
                new_image_dataset,
                maps,
                residuals=compute_residuals(new_image_dataset, dataset, indices),
            )
            dicttonx(nxdict, output_filename)

        self.outputs.dataset = Dataset(
            dataset=new_image_dataset,
            indices=indices,
            bg_indices=inputs.dataset.bg_indices,
            bg_dataset=inputs.dataset.bg_dataset,
        )
        self.outputs.maps = maps



from __future__ import annotations

from typing import Tuple

import numpy

from ..dtypes import Dataset


def apply_shift(
    input_dataset: Dataset,
    shift: Tuple[float, float] | numpy.ndarray,
    selected_axis: int | None = None,
):
    if not isinstance(shift, numpy.ndarray):
        shift = numpy.array(shift)

    dataset = input_dataset.dataset
    indices = input_dataset.indices

    dataset.apply_shift(shift, axis=selected_axis, indices=indices)



import math
from typing import List
from typing import Sequence
from typing import Tuple


def find_linspace_parameters(
    values: Sequence[float], tolerance: float
) -> Tuple[int, float, float]:
    """
    Sets the unique values of the dimension. If the size of the dimension is fixed,
    it automatically sets the first size values, else it finds the unique values.

    :param array_like values: list of values.

    :param tolerance: Tolerance to find the unique values

    :returns linspace parameters: A tuple with dimension size, start and stop

    """
    unique_values = _find_unique_values(values, tolerance)

    return len(unique_values), values[0], values[-1]


def _find_unique_values(values: Sequence[float], tolerance: float) -> List[float]:
    """
    Function that compares the values passed as parameter and returns only the unique
    ones given the dimension's tolerance.

    :param array_like values: list of values to compare.

    :param tolerance: Tolerance to find the unique values

    :returns : An array with uniques values with right dimension size
    """
    unique_values = []

    for val in values:
        if not unique_values:
            unique_values.append(val)
        else:
            unique = True
            for unique_value in unique_values:
                if math.isclose(unique_value, val, rel_tol=tolerance):
                    unique = False
                    break
            if unique:
                unique_values.append(val)
    return unique_values



import os

import h5py
import numpy
import pytest
from ewoks import load_graph
from ewoks import save_graph
from ewokscore import execute_graph
from ewoksorange.bindings import ows_to_ewoks

from darfix.core.grainplot import MomentType
from darfix.tasks.blindsourceseparation import BlindSourceSeparation
from darfix.tasks.blindsourceseparation import Method
from darfix.tasks.dimensiondefinition import DimensionDefinition
from darfix.tasks.grainplot import GrainPlot
from darfix.tasks.hdf5_data_selection import HDF5DataSelection
from darfix.tasks.noiseremoval import NoiseRemoval
from darfix.tasks.roi import RoiSelection
from darfix.tasks.shiftcorrection import ShiftCorrection

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files

from ..utils import generate_ewoks_task_inputs


def test_darfix_example2_hdf5(tmpdir):
    from orangecontrib.darfix import tutorials

    filename = resource_files(tutorials).joinpath("darfix_example_hdf.ows")

    hdf5_dataset_file = resource_files(tutorials).joinpath(
        "hdf5_dataset", "strain.hdf5"
    )
    assert os.path.exists(str(hdf5_dataset_file))
    inputs = [
        *generate_ewoks_task_inputs(
            HDF5DataSelection,
            raw_detector_data_path="/1.1/instrument/my_detector/data",
            raw_input_file=str(hdf5_dataset_file),
            raw_metadata_path="/1.1/instrument/positioners",
            treated_data_dir=str(tmpdir),
        ),
        *generate_ewoks_task_inputs(ShiftCorrection, shift=[0.1, 2]),
    ]
    graph = load_graph(str(filename), inputs=inputs)

    results = graph.execute(output_tasks=True)
    for node_id, task in results.items():
        assert task.succeeded, node_id


def get_inputs(input_filename: str):
    ds_inputs = generate_ewoks_task_inputs(
        HDF5DataSelection,
        raw_input_file=input_filename,
        raw_detector_data_path="/2.1/instrument/my_detector/data",
        raw_metadata_path="/2.1/instrument/positioners",
    )
    dim_inputs = generate_ewoks_task_inputs(
        DimensionDefinition,
        dims={
            0: {"name": "diffry", "kind": 2, "size": 8, "tolerance": 1e-09},
            1: {"name": "diffrx", "kind": 2, "size": 9, "tolerance": 1e-09},
        },
    )
    roi_inputs = generate_ewoks_task_inputs(
        RoiSelection, roi_origin=[198, 114], roi_size=[59, 133]
    )
    noise_inputs = generate_ewoks_task_inputs(
        NoiseRemoval,
        method="median",
        background_type="Data",
        bottom_threshold=0,
        kernel_size=3,
    )
    shiftcorr_inputs = generate_ewoks_task_inputs(ShiftCorrection, shift=[0.0, 0])

    return [*ds_inputs, *dim_inputs, *roi_inputs, *noise_inputs, *shiftcorr_inputs]


# TODO: Fix RockingCurves: it hangs at dataset.apply_fit
@pytest.mark.skip(reason="RockingCurves takes way too long")
def test_example_workflow1(tmpdir, silx_resources):
    """Execute workflow after converting it to an ewoks workflow"""
    ref_filename = silx_resources.getfile("reference_maps.h5")

    from orangecontrib.darfix import tutorials

    filename = resource_files(tutorials).joinpath("darfix_example1.ows")

    graph = ows_to_ewoks(filename)
    input_filename = silx_resources.getfile("input.h5")
    output_filename = str(tmpdir / "maps.h5")
    inputs = [
        *get_inputs(input_filename),
        *generate_ewoks_task_inputs(GrainPlot, filename=output_filename),
        *generate_ewoks_task_inputs(BlindSourceSeparation, method=Method.NICA),
    ]
    positioners = ("diffrx", "diffry")

    execute_graph(graph, inputs=inputs, outputs=[{"all": True}], merge_outputs=False)

    with h5py.File(ref_filename, "r") as ref_file:
        with h5py.File(output_filename, "r") as output_file:
            ref_entry = ref_file["entry"]
            output_entry = output_file["entry"]
            assert list(output_entry.keys()) == [
                "Mosaicity",
                "Orientation distribution",
                *positioners,
            ]

            for pos in positioners:
                for moment in MomentType.values():
                    numpy.testing.assert_allclose(
                        ref_entry[pos][moment][moment],
                        output_entry[pos][moment][moment],
                    )

            numpy.testing.assert_allclose(
                ref_entry["Mosaicity/Mosaicity"], output_entry["Mosaicity/Mosaicity"]
            )
            numpy.testing.assert_allclose(
                ref_entry["Orientation distribution/key/image"],
                output_entry["Orientation distribution/key/image"],
            )


@pytest.mark.parametrize("load_from_json", (True, False))
def test_example_workflow2(tmpdir, load_from_json, silx_resources):
    """Execute workflow after converting it to an ewoks workflow"""
    ref_filename = silx_resources.getfile("reference_maps.h5")

    from orangecontrib.darfix import tutorials

    input_filename = silx_resources.getfile("input.h5")
    output_filename = str(tmpdir / "maps.h5")
    inputs = [
        *get_inputs(input_filename),
        *generate_ewoks_task_inputs(GrainPlot, filename=output_filename),
    ]
    graph = ows_to_ewoks(
        resource_files(tutorials).joinpath("darfix_example_hdf.ows"), inputs=inputs
    )

    if load_from_json:
        json_filename = save_graph(
            graph, tmpdir / "darfix_example_hdf.json", representation="json"
        )
        graph = load_graph(str(json_filename))

    execute_graph(graph, outputs=[{"all": True}], merge_outputs=False)

    positioners = ("diffrx", "diffry")
    with h5py.File(ref_filename, "r") as ref_file:
        with h5py.File(output_filename, "r") as output_file:
            ref_entry = ref_file["entry"]
            output_entry = output_file["entry"]
            assert list(output_entry.keys()) == [
                "Mosaicity",
                "Orientation distribution",
                *positioners,
            ]

            for pos in positioners:
                for moment in MomentType.values():
                    numpy.testing.assert_allclose(
                        ref_entry[pos][moment][moment],
                        output_entry[pos][moment][moment],
                    )

            numpy.testing.assert_allclose(
                ref_entry["Mosaicity/Mosaicity"], output_entry["Mosaicity/Mosaicity"]
            )
            numpy.testing.assert_allclose(
                ref_entry["Orientation distribution/key/image"],
                output_entry["Orientation distribution/key/image"],
            )



from ewoksorange.tests.conftest import qtapp  # noqa F401

from darfix.gui.utils.data_path_selection import DataPathSelection
from darfix.tests.utils import create_scans


def test_DataPathSelection(tmp_path, qtapp):  # noqa F811
    """Test DataPathSelection pattern"""

    default_detector_data_path = r"{scan}/measurement/{detector}"
    default_positioners_data_path = r"{scan}/instrument/positioners"

    test_folder = tmp_path / "test_data_path_selection"
    test_folder.mkdir()
    input_file = test_folder / "scans.hdf5"

    create_scans(file_path=input_file)

    # test the find_detector_dataset callback
    widget = DataPathSelection(
        pattern=default_detector_data_path,
    )

    widget.setInputFile(str(input_file))
    qtapp.processEvents()

    assert widget.getPattern() == default_detector_data_path
    assert widget.getDefaultPattern() == default_detector_data_path
    assert widget._dataPathFinder._solved_pattern == "{scan}/measurement/my_detector"
    assert widget.getExample() == "/1.1/measurement/my_detector"
    assert widget._dataPathFinder._can_be_solved is True

    # test with find_positioner callback
    widget.setPattern(
        pattern=default_positioners_data_path,
    )
    qtapp.processEvents()
    assert widget.getPattern() == default_positioners_data_path
    assert widget.getExample() == "/1.1/instrument/positioners"
    assert widget.getDefaultPattern() == default_detector_data_path
    widget.setDefaultPattern(default_positioners_data_path)
    assert widget.getDefaultPattern() == default_positioners_data_path
    assert widget._dataPathFinder._solved_pattern == "{scan}/instrument/positioners"
    assert widget._dataPathFinder._can_be_solved is True

    # test with a not existing foler to search
    pattern_not_existing = "{scan}/pattern/not/existing"
    widget.setPattern(
        pattern=pattern_not_existing,
    )
    qtapp.processEvents()
    assert widget.getExample() == ""
    assert widget.getPattern() == pattern_not_existing
    assert widget._dataPathFinder._solved_pattern == pattern_not_existing
    assert widget._dataPathFinder._can_be_solved is False



from __future__ import annotations

import multiprocessing
from functools import partial
from multiprocessing import Pool
from typing import Generator
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy
import tqdm
from silx.utils.enum import Enum as _Enum

from ..io.utils import create_nxdata_dict
from ..processing.rocking_curves import FitMethod
from ..processing.rocking_curves import fit_1d_rocking_curve
from ..processing.rocking_curves import fit_2d_rocking_curve
from ..processing.rocking_curves import fit_3d_rocking_curve
from .utils import NoDimensionsError
from .utils import TooManyDimensionsForRockingCurvesError

Indices = Union[range, numpy.ndarray]


class Maps_1D(_Enum):
    """Names of the fitting parameters of the 1D fit result. Each result is a map of frame size."""

    AMPLITUDE = "Amplitude"
    FWHM = "FWHM"
    PEAK = "Peak position"
    BACKGROUND = "Background"


class Maps_2D(_Enum):
    """Names of the fitting parameters of the 2D fit result. Each result is a map of frame size."""

    AMPLITUDE = "Amplitude"
    PEAK_X = "Peak position first motor"
    PEAK_Y = "Peak position second motor"
    FWHM_X = "FWHM first motor"
    FWHM_Y = "FWHM second motor"
    BACKGROUND = "Background"
    CORRELATION = "Correlation"


class Maps_3D(_Enum):
    """Names of the fitting parameters of the 3D fit result. Each result is a map of frame size."""

    AMPLITUDE = "Amplitude"
    PEAK_X = "Peak position first motor"
    PEAK_Y = "Peak position second motor"
    PEAK_Z = "Peak position third motor"
    FWHM_X = "FWHM first motor"
    FWHM_Y = "FWHM second motor"
    FWHM_Z = "FWHM third motor"
    BACKGROUND = "Background"
    CORRELATION_XY = "Cross-correlation between first and second motors"
    CORRELATION_XZ = "Cross-correlation between first and third motors"
    CORRELATION_YZ = "Cross-correlation between second and third motors"


MAPS_1D: Tuple[Maps_1D] = Maps_1D.values()
MAPS_2D: Tuple[Maps_2D] = Maps_2D.values()
MAPS_3D: Tuple[Maps_3D] = Maps_3D.values()


def _rocking_curves_per_px(
    data: numpy.ndarray, moments: numpy.ndarray | None = None, indices=None
) -> (
    Generator[Tuple[numpy.ndarray, None], None, None]
    | Generator[Tuple[numpy.ndarray, numpy.ndarray], None, None]
):
    """
    Generator that returns the rocking curve for every pixel

    :param ndarray data: data to analyse
    :param moments: array of same shape as data with the moments values per pixel and image, optional
    :type moments: Union[None, ndarray]
    """
    for i in range(data.shape[1]):
        for j in range(data.shape[2]):
            if indices is None:
                new_data = data[:, i, j]
            else:
                new_data = numpy.zeros(data.shape[0])
                new_data[indices] = data[indices, i, j]
            if moments is not None:
                yield new_data, moments[:, i, j]
            yield new_data, None


def fit_1d_data(
    data: numpy.ndarray,
    moments: numpy.ndarray | None = None,
    values: List[numpy.ndarray] | numpy.ndarray | None = None,
    int_thresh: float = 15.0,
    method: FitMethod | None = None,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Fit data in axis 0 of data"""

    rocking_curves_per_px = _rocking_curves_per_px(data, moments)
    cpus = multiprocessing.cpu_count()
    curves, maps = [], []
    with Pool(cpus - 1) as p:
        for curve, pars in tqdm.tqdm(
            p.imap(
                partial(
                    fit_1d_rocking_curve,
                    x_values=values,
                    int_thresh=int_thresh,
                    method=method,
                ),
                rocking_curves_per_px,
            ),
            total=data.shape[1] * data.shape[2],
        ):
            curves.append(list(curve))
            maps.append(list(pars))

    return numpy.array(curves).T.reshape(data.shape), numpy.array(maps).T.reshape(
        (len(MAPS_1D), data.shape[-2], data.shape[-1])
    )


def fit_2d_data(
    data: numpy.ndarray,
    values: List[numpy.ndarray] | numpy.ndarray,
    shape: Tuple[int, int],
    moments: numpy.ndarray | None = None,
    int_thresh: float | None = 15.0,
    indices: Indices | None = None,
    method: FitMethod | None = None,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Fit data in axis 0 of data"""
    rocking_curves_per_px = _rocking_curves_per_px(data, moments, indices)
    curves, maps = [], []
    cpus = multiprocessing.cpu_count()
    with Pool(cpus - 1) as p:
        for curve, pars in tqdm.tqdm(
            p.imap(
                partial(
                    fit_2d_rocking_curve,
                    x_values=values,
                    shape=shape,
                    int_thresh=int_thresh,
                    method=method,
                ),
                rocking_curves_per_px,
            ),
            total=data.shape[-2] * data.shape[-1],
        ):
            curves.append(list(curve))
            maps.append(list(pars))

    curves = numpy.array(curves).T
    if indices is not None:
        curves = curves[indices]
        data_shape = data[indices].shape
    else:
        data_shape = data.shape

    return curves.reshape(data_shape), numpy.array(maps).T.reshape(
        (len(MAPS_2D), data.shape[-2], data.shape[-1])
    )


def fit_3d_data(
    data: numpy.ndarray,
    motor_values: Sequence[numpy.ndarray] | numpy.ndarray,
    shape: Tuple[int, int, int],
    moments: numpy.ndarray | None = None,
    int_thresh: float | None = 15.0,
    indices: Indices | None = None,
    method: FitMethod | None = None,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Fit the rocking curves using a trivariate gaussian functions

    :param data: Stack of images (Nframes, Npx, Npy)
    :param motor_values: Values of the 3 motor positions. The order is (fast, mid, slow). Shape shoudl be (3, Nframes).
    :param shape: Number of steps in each motor direction. The order is (fast, mid, slow). The product of these number should equal Nframes.
    :param moments: Moments of the data to be used as starting point for the fit.
    :param int_thresh: Minimum intensity threshold. If the summed intensity is below this value, the frame will be skipped. Defaults to 15.0.
    :param indices: Frames for which the fit should be made. Defaults to None (all frames).
    :param method: Fit method. 'lm', 'trf' or 'dogbox'. Defaults to 'trf'.
    :returns: A tuple containing the fitted curves (Nframes, Npx, Npy) and the fit parameters (Nparams, Npx, Npy).
    """
    rocking_curves_per_px = _rocking_curves_per_px(data, moments, indices)
    if isinstance(motor_values, Sequence):
        motor_values = numpy.array(motor_values)
    if not isinstance(motor_values, numpy.ndarray):
        raise TypeError(
            f"motor_values should be a sequence or a numpy array. Got {type(motor_values)}"
        )

    curves, maps = [], []
    for rocking_curve in rocking_curves_per_px:
        curve, fit_params = fit_3d_rocking_curve(
            rocking_curve, motor_values, shape, int_thresh, method
        )
        curves.append(list(curve))
        maps.append(list(fit_params))

    # Arrays have dimensions (Npixels, Ncomponents): transpose to have the pixels as last dimension
    curves = numpy.array(curves).T
    maps = numpy.array(maps).T
    # And reshape to fit the data:
    # curves will be (Nframes, Npx, Npy)
    # maps will be (Nparams, Npx, Npy)
    if indices is not None:
        curves = curves[indices].reshape(data[indices].shape)
    else:
        curves = curves.reshape(data.shape)

    return curves, maps.reshape((len(MAPS_3D), data.shape[-2], data.shape[-1]))


def generate_rocking_curves_nxdict(
    dataset,  # ImageDataset. Cannot type due to circular import
    maps: numpy.ndarray,
    residuals: numpy.ndarray | None,
) -> dict:
    if not dataset.dims.ndim:
        raise NoDimensionsError("generate_rocking_curves_nxdict")
    entry = "entry"

    nx = {
        entry: {"@NX_class": "NXentry"},
        "@NX_class": "NXroot",
        "@default": entry,
    }

    if dataset.transformation:
        axes = [
            dataset.transformation.yregular,
            dataset.transformation.xregular,
        ]
        axes_names = ["y", "x"]
        axes_long_names = [
            dataset.transformation.label,
            dataset.transformation.label,
        ]
    else:
        axes = None
        axes_names = None
        axes_long_names = None

    if dataset.dims.ndim == 1:
        map_names = MAPS_1D
    elif dataset.dims.ndim == 2:
        map_names = MAPS_2D
    else:
        raise TooManyDimensionsForRockingCurvesError()

    for i, map_name in enumerate(map_names):
        signal = maps[i]
        nx[entry][map_name] = create_nxdata_dict(
            signal, map_name, axes, axes_names, axes_long_names
        )
    if residuals is not None:
        nx[entry]["Residuals"] = create_nxdata_dict(
            residuals, "Residuals", axes, axes_names, axes_long_names
        )
    nx[entry]["@default"] = Maps_1D.AMPLITUDE.value

    return nx


def compute_residuals(
    target_dataset,  # ImageDataset. Cannot type due to circular import
    original_dataset,  # ImageDataset. Cannot type due to circular import
    indices: numpy.ndarray | None,
):
    return numpy.sqrt(
        numpy.subtract(target_dataset.zsum(indices), original_dataset.zsum(indices))
        ** 2
    )



from __future__ import annotations

from ewokscore.missing_data import is_missing_data
from silx.gui import qt

from darfix import dtypes
from darfix.core.noiseremoval import NoiseRemovalOperation
from darfix.gui.noiseremoval.noiseRemovalWidget import NoiseRemovalWidget
from darfix.tasks.noiseremoval import NoiseRemoval
from orangecontrib.darfix.widgets.operationBase import OperationWidgetBase


class NoiseRemovalWidgetOW(OperationWidgetBase, ewokstaskclass=NoiseRemoval):
    name = "noise removal"
    description = "A widget to perform various noise removal operations"
    icon = "icons/noise_removal.png"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("operations",)

    def __init__(self):
        super().__init__()

    def createMainWidget(self, inputs: dict) -> NoiseRemovalWidget:
        widget = NoiseRemovalWidget()
        widget.sigLaunchOperation.connect(self._execute_noise_removal_operation)
        incomingOperations = inputs.get("operations", [])
        widget.setIncomingOperations(incomingOperations)
        widget.setDefaultParameters(incomingOperations)
        return widget

    def saveMainWidget(self):
        self.set_default_input(
            "operations",
            list(
                filter(
                    lambda op: not op.get("aborted", False),
                    self.mainWidget.getOperationHistory(),
                )
            ),
        )

    @property
    def mainWidget(self) -> NoiseRemovalWidget:
        return super().mainWidget

    def handleNewSignals(self) -> None:
        super().handleNewSignals()
        dataset = self.get_task_input_value("dataset")
        self.setDataset(dataset, pop_up=True)

    def setDataset(self, dataset: dtypes.Dataset | None, pop_up=True):
        if dataset is None:
            return
        self.mainWidget.setDataset(dataset)

        if self.mainWidget.hasIncomingOperations() and not is_missing_data(
            self.get_task_input_value("dataset")
        ):
            messagebox = qt.QMessageBox(
                qt.QMessageBox.Icon.Question,
                "Replay Operations ?",
                "Do you want to apply the following operations from last save now ?\n IF NOT, the history of operations will be cleared.\n\n"
                + "\n\n".join(
                    [
                        f"{i} - {str(ope)}"
                        for i, ope in enumerate(self.mainWidget.getIncomingOperations())
                    ]
                ),
                buttons=qt.QMessageBox.StandardButton.Yes
                | qt.QMessageBox.StandardButton.No,
            )
            ret = messagebox.exec()
            if ret == qt.QMessageBox.StandardButton.Yes:
                self.execute_ewoks_task_without_propagation()

        if pop_up:
            self.open()

    def _execute_noise_removal_operation(
        self, operation: NoiseRemovalOperation
    ) -> None:
        incomingOperations = [operation]
        self.mainWidget.setIncomingOperations(incomingOperations)
        self.set_dynamic_input("operations", incomingOperations)

        self.execute_ewoks_task_without_propagation()

    def task_output_changed(self):
        if self.isAborted() and self.replayAfterAbort():
            return
        if self.isAborted():
            self.mainWidget.abortIncomingOperations()
        else:
            self.mainWidget.extendOperationHistory()

        super().task_output_changed()
        self.mainWidget.refreshPlot()

    def replayAfterAbort(self) -> bool:
        """Create a messagebox to ask user if he wants to replay after abort. Return True if Yes."""
        messagebox = qt.QMessageBox(
            qt.QMessageBox.Icon.Question,
            "Operation aborted !",
            "Operation(s) was not applied to the whole stack.\n"
            "Do you want to replay all previous operations to recover last state ?\n"
            "If not, the stack will stay in a partially modified state. You can still reset dataset if you want to restart the noise removal process.\n",
            buttons=qt.QMessageBox.StandardButton.Yes
            | qt.QMessageBox.StandardButton.No,
        )

        replay = messagebox.exec() == qt.QMessageBox.StandardButton.Yes

        if replay:
            ops = self.mainWidget.getOperationHistory()
            self.reset()
            self.mainWidget.setIncomingOperations(ops, clear=True)
            self.set_dynamic_input("operations", ops)
            self.execute_ewoks_task_without_propagation()

        return replay



import numpy

from darfix.processing.imageOperations import mask_removal

from .utils import create_1d_dataset


def test_apply_mask_from_file(tmpdir):
    dataset = create_1d_dataset(tmpdir, motor1="diffrx", motor2="diffry")
    mask_shape = dataset.data.shape[1:]
    mask = numpy.ones(mask_shape)
    mask[20:40, 20:40] = 0

    masked_data = dataset.data.copy()
    mask_removal(masked_data, mask)
    # Do the assert frame by frame since we can't slice through DataUrls
    for i in range(len(dataset.data)):
        expected_data = dataset.data[i]
        expected_data[20:40, 20:40] = 0
        numpy.testing.assert_allclose(masked_data[i], expected_data)


def test_apply_mask_from_memory(tmpdir):
    dataset = create_1d_dataset(tmpdir, motor1="diffrx", motor2="diffry")
    mask_shape = dataset.data.shape[1:]
    mask = numpy.ones(mask_shape)
    mask[20:40, 20:40] = 0

    masked_data = dataset.data.copy()
    mask_removal(masked_data, mask)

    expected_data = dataset.data
    expected_data[:, 20:40, 20:40] = 0
    numpy.testing.assert_allclose(masked_data, expected_data)



import importlib.metadata
import sys
from argparse import ArgumentParser

try:
    from ewoksorange.canvas.main import main as ewoksorange_main
except ImportError:
    ewoksorange_main = None


def main(argv=None):
    parser = ArgumentParser(add_help=False)

    parser.add_argument(
        "--version",
        action="store_true",
        help="Display version",
    )
    if argv is None:
        argv = sys.argv
    options, _ = parser.parse_known_args(argv[1:])

    if options.version:
        print(f"Darfix version: {importlib.metadata.version('darfix')}")
        return
    if ewoksorange_main is None:
        raise ImportError("Install darfix[full] to use the Orange canvas.")
    ewoksorange_main()


if __name__ == "__main__":
    sys.exit(main())
